import numpy as np
import matplotlib.pyplot as plt
import math as mt
from itertools import permutations
import numdifftools as nd


def generate_data_multi(n, means, scale, weights, distribution='gaussian', kernel=None):
    """
    Generate n samples from a mixture with r components,
    where each component i has mean means[i] and weight weights[i].
    The parameter 'distribution' can be 'gaussian', 'lorentz', 'uniform' or 'student'.
    For 'student', the degrees of freedom is taken from kernel.nu.
    This version works even when n == 1.
    """

    #means correspond aux tau_i
    #weights correspond aux a_i, qu'on doit normaliser pour avoir une distribution de probabilité
    #scale correspond à l'écart-type, on le fixera à 1


    r = len(means)
    # Special case: n == 1
    if n == 1:
        comp = np.random.choice(r, p=weights)
        if distribution == 'gaussian':
            return np.array([np.random.normal(means[comp], scale)])
        elif distribution == 'lorentz':
            sample = np.random.standard_cauchy() * scale + means[comp]
            return np.array([np.clip(sample, -50, 50)])
        elif distribution == 'uniform':
            return np.array([np.random.uniform(means[comp] - scale, means[comp] + scale)])
        elif distribution == 'student':
            df = kernel.nu if (kernel is not None and hasattr(kernel, 'nu')) else 3
            return np.array([np.random.standard_t(df) * scale + means[comp]])
        else:
            raise ValueError("Unknown distribution type")

    # For n > 1: use a multinomial draw to exactly split n samples.
    counts = np.random.multinomial(n, weights)
    data = []
    for i in range(r):
        n_i = counts[i]
        if distribution == 'gaussian':
            samples = np.random.normal(means[i], scale, n_i)
        elif distribution == 'lorentz':
            samples = np.random.standard_cauchy(n_i) * scale + means[i]
            samples = np.clip(samples, -50, 50)
        elif distribution == 'uniform':
            samples = np.random.uniform(means[i] - scale, means[i] + scale, n_i)
        elif distribution == 'student':
            df = kernel.nu if (kernel is not None and hasattr(kernel, 'nu')) else 3
            samples = np.random.standard_t(df, n_i) * scale + means[i]
        else:
            raise ValueError("Unknown distribution type")
        data.append(samples)
    data = np.concatenate(data)
    np.random.shuffle(data)
    return data



def generate_data(m,taus,weights,sigma=1):
    if sum(weights) !=1:
        s = sum(weights)
        weights = np.array(weights)/s

    #Return m data distributed with the pdf : Sum weights[i]*N(taus[i],sigma)
    r = len(taus)

    quelle_cloche = np.random.choice(r , size = m , p = weights)
    #quelle_cloche[i] correspond à la cloche de provenance de data[i]
 
    #taus[quelle_cloche][i] donne la moyenne de la gaussienne dont provient data[i]
    data = np.random.normal(np.array(taus)[quelle_cloche],sigma)
    return data
        




def genere_hist(weights,means):
   
    data = generate_data(100000,means,weights)
    plt.hist(data, bins=50, density=True, alpha=0.6, color='b')
    plt.xlabel("Valeurs")
    plt.ylabel("Densité")
    plt.title("Histogramme d'une somme de lois normales")
    plt.show()



def FCE(data, freqs):
    # Compute the empirical characteristic function at frequencies in freqs.
    #L'idée est de confondre les données par leur fonction caractéristique empirique prise en N fréquences.
    #Cela réduit la taille du problème de m à N.

    #L'utilisation de np notamment pour np.exp vectorise les calculs -> gains de temps et gère les nombres complexes

    phi = np.array([np.mean(np.exp(-2j * np.pi * f * data)) for f in freqs])

    return phi.reshape(-1, 1) #Utile pour l'avoir en tant que vecteur pour après


def sample_uniform_frequencies(n):
    #N = 2n+1 est la nouvelle taille du problème
    N = 2 * n + 1
    return np.arange(-n, n + 1).reshape(-1, 1) / N


def fourier(x,sigma=1):

    return np.exp(-2 * (x * sigma * np.pi)**2)



def ESPRIT_(x, r, my_fourier, freqs1, true_freqs):

    x_vec = x.flatten()
    N = len(x_vec)
    L = N // 2
    M = N - L + 1

    X = np.zeros((L, M), dtype=complex)
    for i in range(L):
        X[i, :] = x_vec[i:i+M]

    U, s, Vh = np.linalg.svd(X, full_matrices=False)
    U_r = U[:, :r]

    U1 = U_r[:-1, :]
    U2 = U_r[1:, :]

    Psi = np.linalg.pinv(U1) @ U2
    eigvals, _ = np.linalg.eig(Psi)

    f_est = N * np.angle(eigvals) / (2 * np.pi)
    t0_est = -np.sort(f_est)

    V0 = my_fourier(true_freqs).reshape(-1, 1) * np.exp(2j * np.pi * np.outer(freqs1.flatten(), t0_est))
    a0 = np.linalg.lstsq(V0, x_vec, rcond=None)[0]

    return np.concatenate((a0, t0_est))


def S_err(theta_est, r, theta_star):
    # Assumes theta_est and theta_star are concatenations: [amplitudes, centroids]
    centroids_est = np.sort(theta_est[r:])
    centroids_true = np.sort(theta_star[r:])
    return np.linalg.norm(centroids_est - centroids_true)



def permute_and_compute_norm(a, b, k):
    # Diviser a et b en 2 parties : weights_a (premières k composantes) et means_tau (les autres)
    a_means, a_weights = a[k:], a[:k]

    # Permuter les parties means dans a, et pareil pour les parties weights dans a
    a_means_permuted= list(permutations(a_means))
    a_weights_permuted = list(permutations(a_weights))

    min_norm = np.inf  # Initialiser la norme minimale à un grand nombre

    # Initialiser best_perm à la permutation identité
    best_perm = (tuple(a_weights), tuple(a_means))

    # Tester toutes les permutations de means et weights
    for perm_means_a in a_means_permuted:
        for perm_weights_a in a_weights_permuted:
                
            # Recomposer les vecteurs permutés
            a_permuted = np.concatenate([np.array(perm_weights_a), np.array(perm_means_a)])
                    
                    
            # Calculer la norme Euclidienne
            norm = np.linalg.norm(a_permuted - b)
            if norm < min_norm:
                min_norm = norm
                best_perm = (perm_weights_a, perm_means_a)

    # Maintenant que nous avons best_perm, appliquer la permutation à a (modification directe de a)
    
    perm_weights_a, perm_means_a = best_perm

    # Appliquer la permutation sur les "means" et "weights" de a
    a_means_permuted = np.array(perm_means_a)
    a_weights_permuted = np.array(perm_weights_a)

    # Recomposer a avec les nouvelles parties permutées
    c = np.concatenate([a_weights_permuted,a_means_permuted])

    return min_norm , c # Retourner la norme minimale trouvée



def batch_algo_gradient_descent(m, n, means, weights, num_steps=100, eta=0.1):
    """
    First, perform an ESPRIT estimation using a batch of m samples.
    Then, run a batch (offline) preconditioned gradient descent refinement for num_steps,
    using the fixed empirical characteristic function computed from the m samples.
    n is here for the frequencies : N=2n+1 frequencies used.
    """
    r = len(means)


    # Generate batch data and compute its FCE
    data = generate_data(m, means, weights)
    freqs = sample_uniform_frequencies(n)  # shape (N_freq, 1)
    baseline_freqs = sample_uniform_frequencies(n)


    # Compute the empirical characteristic function once for the batch
    Y = FCE(data, freqs)  # shape (N_freq, 1)

    # Initial ESPRIT estimation: theta_est = [a_est (r,), tau_est (r,)]
    theta_est = ESPRIT_(Y, r, fourier, baseline_freqs, freqs)
    a_est = theta_est[:r].reshape(-1, 1)
    tau_est = theta_est[r:]

    print(theta_est)

    # Precompute the "G" matrix: diagonal with entries given by kernel's Fourier transform at freqs.
    G_diag = fourier(freqs).flatten()  # shape (N_freq,)
    G_mat = np.diag(G_diag)  # shape (N_freq, N_freq)

    # Dictionary function Phi: for a given tau vector (r,), returns a matrix (N_freq x r)
    def Phi(tau):
        return np.exp(-2j * np.pi * np.outer(freqs.flatten(), tau))

    # Model function: Y_model = G * Phi(tau) * a.
    def Y_model(a, tau):
        return G_mat @ (Phi(tau) @ a)

    # Compute K''(0): for K(t)=sum_f |ĝ(f)|^2 exp(2i*pi*f*t), we have
    # K''(0)= - (2*pi)^2 * sum_f |ĝ(f)|^2 * f^2.
    f_vals = freqs.flatten()
    Kpp0 = - (2 * np.pi)**2 * np.sum(np.abs(fourier(freqs).flatten())**2 * (f_vals**2))

    # Run batch gradient descent updates on the fixed FCE Y
    for step in range(num_steps):
        # Compute the current model prediction
        Y_estimated = Y_model(a_est, tau_est)
        # Residual: difference between model prediction and observed FCE
        r_vec = Y_estimated - Y  # shape (N_freq, 1)

        # Gradient with respect to amplitudes: grad_a = (G*Phi(tau))^H * r
        grad_a = (G_mat @ Phi(tau_est)).conj().T @ r_vec  # shape (r, 1)

        # Gradient with respect to tau:
        grad_tau = np.zeros_like(tau_est, dtype=complex)
        for i in range(r):
            dPhi_dtau = 2j * np.pi * freqs.flatten() * np.exp(2j * np.pi * freqs.flatten() * tau_est[i])
            grad_tau[i] = np.real(a_est[i, 0] * ((G_mat @ dPhi_dtau).conj().T @ r_vec)[0])

        # Preconditioner for tau-update:
        #P_tau = (1/(-Kpp0)) / (np.abs(a_est.flatten())**2)

        P_tau=1
        # Gradient descent update with learning rate eta
        a_est = a_est - eta * grad_a
        tau_est = tau_est - eta * (P_tau * grad_tau)

    # Final estimated parameters (concatenated amplitudes and centroids)
    theta_est_final = np.concatenate((a_est.flatten(), tau_est))
    # True parameters vector: [weights, means]
    theta_star = np.concatenate((weights, means))

    norm , theta_est_final_prime = permute_and_compute_norm(theta_est_final,theta_star,r)
    return norm , theta_est_final_prime , theta_star 



def batch_algo_gradient_descent_without_estimator(m, n, means, weights,est, num_steps=100, eta=0.1):
    """
    First, perform an ESPRIT estimation using a batch of m samples.
    Then, run a batch (offline) preconditioned gradient descent refinement for num_steps,
    using the fixed empirical characteristic function computed from the m samples.
    n is here for the frequencies : N=2n+1 frequencies used.
    """
    r = len(means)


    # Generate batch data and compute its FCE
    data = generate_data(m, means, weights)
    freqs = sample_uniform_frequencies(n)  # shape (N_freq, 1)
    


    # Compute the empirical characteristic function once for the batch
    Y = FCE(data, freqs)  # shape (N_freq, 1)

    est = np.array(est)
    a_est = est[:r].reshape(-1, 1)
    tau_est = est[r:]

    
    
    # Precompute the "G" matrix: diagonal with entries given by kernel's Fourier transform at freqs.
    G_diag = fourier(freqs).flatten()  # shape (N_freq,)
    G_mat = np.diag(G_diag)  # shape (N_freq, N_freq)

    # Dictionary function Phi: for a given tau vector (r,), returns a matrix (N_freq x r)
    def Phi(tau):
        return np.exp(-2j * np.pi * np.outer(freqs.flatten(), tau))

    # Model function: Y_model = G * Phi(tau) * a.
    def Y_model(a, tau):
        return G_mat @ (Phi(tau) @ a)

    # Compute K''(0): for K(t)=sum_f |ĝ(f)|^2 exp(2i*pi*f*t), we have
    # K''(0)= - (2*pi)^2 * sum_f |ĝ(f)|^2 * f^2.
    f_vals = freqs.flatten()
    Kpp0 = - (2 * np.pi)**2 * np.sum(np.abs(fourier(freqs).flatten())**2 * (f_vals**2))

    # Run batch gradient descent updates on the fixed FCE Y
    for step in range(num_steps):
        # Compute the current model prediction
        Y_estimated = Y_model(a_est, tau_est)
        # Residual: difference between model prediction and observed FCE
        r_vec = Y_estimated - Y  # shape (N_freq, 1)

        # Gradient with respect to amplitudes: grad_a = (G*Phi(tau))^H * r
        grad_a = (G_mat @ Phi(tau_est)).conj().T @ r_vec  # shape (r, 1)

        # Gradient with respect to tau:
        grad_tau = np.zeros_like(tau_est, dtype=complex)
        for i in range(r):
            dPhi_dtau = 2j * np.pi * freqs.flatten() * np.exp(2j * np.pi * freqs.flatten() * tau_est[i])
            grad_tau[i] = np.real(a_est[i, 0] * ((G_mat @ dPhi_dtau).conj().T @ r_vec)[0])

        # Preconditioner for tau-update:
        #P_tau = (1/(-Kpp0)) / (np.abs(a_est.flatten())**2)

        P_tau=1
        # Gradient descent update with learning rate eta
        a_est = a_est - eta * grad_a
        tau_est = tau_est - eta * (P_tau * grad_tau)

    # Final estimated parameters (concatenated amplitudes and centroids)
    theta_est_final = np.concatenate((a_est.flatten(), tau_est))
    # True parameters vector: [weights, means]
    theta_star = np.concatenate((weights, means))

    norm , theta_est_final_prime = permute_and_compute_norm(theta_est_final,theta_star,r)
    return norm , theta_est_final_prime , theta_star 






def loss(teta,teta_star,r,freqs,):
    a , a_star= np.array(teta[:r]).reshape(-1, 1) , np.array(teta_star[:r]).reshape(-1, 1)
    tau , tau_star = teta[r:] , teta_star[r:]


    # Precompute the "G" matrix: diagonal with entries given by kernel's Fourier transform at freqs
    G_diag = fourier(freqs).flatten()  # shape (N_freq,)
    G_mat = np.diag(G_diag)  # shape (N_freq, N_freq)


    # Dictionary function Phi: for a given tau vector (r,), returns a matrix (N_freq x r)
    def Phi(tau):

        return np.exp(2j * np.pi * np.outer(freqs.flatten(), tau))



    # Model function: Y_model = G * Phi(tau) * a.
    def Y_model(a, tau):
        return G_mat @ (Phi(tau) @ a)

    Y = Y_model(a,tau)
    Y_star = Y_model(a_star,tau_star)

    return 1/2 * np.linalg.norm(Y-Y_star)**2

def show_loss(tau_star, tau_min,tau_max, nb_values,n): #tau == means
    """
    We take r=2, and let a1 = a2 = 1/2.

    We want to look at the landscape of the loss function L(teta) in this configuration.
    We Here do uniform sampling with N=2n+1 centered frequencies.
    """

    assert len(tau_star) == 2 , "Issue : r !=2 "

    tau1_vals = np.linspace(tau_min,tau_max,nb_values)
    tau2_vals = np.linspace(tau_min,tau_max,nb_values)

    tau1 , tau2 = np.meshgrid(tau1_vals,tau2_vals) #Création de la grille

    freqs=sample_uniform_frequencies(n)
     
    # Vectorized version of loss function
    def f_vectorized(tau1_vals, tau2_vals):
        # Crée une grille de tau1 et tau2 en une seule opération
        teta_vals = np.array([1/2, 1/2])  # a1 = a2 = 1/2
        
        # Créer la grille de teta avec les bonnes valeurs
        teta = np.vstack([np.repeat(teta_vals[0], tau1_vals.size), 
                          np.repeat(teta_vals[1], tau2_vals.size),
                          tau1_vals.flatten(), tau2_vals.flatten()]).T
        
        teta_star = np.concatenate([np.array([1/2, 1/2]), tau_star])
        
        # Calcul vectorisé de la perte pour toute la grille de tau1 et tau2
        losses = np.array([loss(teta_row, teta_star, 2, freqs) for teta_row in teta])
        return losses.reshape(tau1_vals.shape)

    Z = f_vectorized(tau1,tau2)

    # Tracé de la carte de chaleur
    plt.figure(figsize=(8, 6))
    plt.contourf(tau1, tau2, Z, levels=50, cmap='viridis')
    plt.colorbar(label="f(tau1, tau2)")
    plt.xlabel("Tau1")
    plt.ylabel("Tau2")
    plt.title("Heatmap of loss function.")

    plt.show()






def show_loss_path(tau_est,tau_star,tau_min,tau_max,nb_values,m,n,num_steps =100,eta=0.1):
    """
    We take r=2, and let a1 = a2 = 1/2.

    We want to look at the landscape of the loss function L(teta) in this configuration.
    We Here do uniform sampling with N=2n+1 centered frequencies.

    Then, we do a gradient descent from theta_est to (we hope) theta_star.
    """
    assert len(tau_star) == 2 , "Issue : r !=2 "

    tau1_vals = np.linspace(tau_min,tau_max,nb_values)
    tau2_vals = np.linspace(tau_min,tau_max,nb_values)

    tau1 , tau2 = np.meshgrid(tau1_vals,tau2_vals) #Création de la grille

    freqs=sample_uniform_frequencies(n)
     
    # Vectorized version of loss function
    def f_vectorized(tau1_vals, tau2_vals):
        # Crée une grille de tau1 et tau2 en une seule opération
        teta_vals = np.array([1/2, 1/2])  # a1 = a2 = 1/2
        
        # Créer la grille de teta avec les bonnes valeurs
        teta = np.vstack([np.repeat(teta_vals[0], tau1_vals.size), 
                          np.repeat(teta_vals[1], tau2_vals.size),
                          tau1_vals.flatten(), tau2_vals.flatten()]).T
        
        teta_star = np.concatenate([np.array([1/2, 1/2]), np.array(tau_star)])
        
        # Calcul vectorisé de la perte pour toute la grille de tau1 et tau2
        losses = np.array([loss(teta_row, teta_star, 2, freqs) for teta_row in teta])
        return losses.reshape(tau1_vals.shape)

    Z = f_vectorized(tau1,tau2)

    # Tracé de la carte de chaleur
    plt.figure(figsize=(8, 6))
    plt.contourf(tau1, tau2, Z, levels=50, cmap='viridis')
    plt.colorbar(label="f(tau1, tau2)")
    plt.xlabel("Tau1")
    plt.ylabel("Tau2")
    plt.title("Heatmap of loss function.")

    """
    Maintenant on fait le batch_algo_gradient_descent en gardant en mémoire les teta_k
    """
    
    means = tau_star
    weights = [1/2,1/2]
    r = len(means)
    N=2*n+1
 
    # Generate batch data and compute its FCE
    data = generate_data(m, means, weights)
    freqs = sample_uniform_frequencies(n)  # shape (N_freq, 1)
    baseline_freqs = sample_uniform_frequencies(n)


    # Compute the empirical characteristic function once for the batch

    #Y = FCE(data, freqs)  # shape (N_freq, 1) #Noise coming from data generation

    
    # Initial ESPRIT estimation: theta_est = [a_est (r,), tau_est (r,)]

    ###theta_est = ESPRIT_(Y, r, fourier, baseline_freqs, freqs)
    #theta_est = np.array([1/2,1/2,4.9,5])
    #theta_est [0], theta_est[1] = 1/2,1/2 #ICI AUCUNE ESTIMATION SUR LES WEIGHTS

    a_est = np.array([1/2,1/2])
    a_star = np.array([1/2,1/2])
    tau_est = np.array(tau_est)

    theta_est = np.concatenate((a_est,tau_est))

    a_est = a_est.reshape(-1, 1)
    a_star = a_star.reshape(-1,1)
    tau1_traj=[tau_est[0]]
    tau2_traj=[tau_est[1]]


    # Precompute the "G" matrix: diagonal with entries given by kernel's Fourier transform at freqs.
    G_diag = fourier(freqs).flatten()  # shape (N_freq,)
    G_mat = np.diag(G_diag)  # shape (N_freq, N_freq)

    # Dictionary function Phi: for a given tau vector (r,), returns a matrix (N_freq x r)
    def Phi(tau):
        return np.exp(-2j * np.pi * np.outer(freqs.flatten(), tau))

    # Model function: Y_model = G * Phi(tau) * a.
    def Y_model(a, tau):
        return G_mat @ (Phi(tau) @ a)

    Y = Y_model(a_star,tau_star) # No noise assumed, perfect data m -> +inf

    # Compute K''(0): for K(t)=sum_f |ĝ(f)|^2 exp(2i*pi*f*t), we have
    # K''(0)= - (2*pi)^2 * sum_f |ĝ(f)|^2 * f^2.
    f_vals = freqs.flatten()
    Kpp0 = - (2 * np.pi)**2 * np.sum(np.abs(fourier(freqs).flatten())**2 * (f_vals**2))
    print(Kpp0)
    # Run batch gradient descent updates on the fixed FCE Y
    for step in range(num_steps):
        # Compute the current model prediction
        
        Y_estimated = Y_model(a_est, tau_est)
        # Residual: difference between model prediction and observed FCE
        r_vec = Y_estimated - Y  # shape (N_freq, 1)

        # Gradient with respect to amplitudes: grad_a = (G*Phi(tau))^H * r
        grad_a = (G_mat @ Phi(tau_est)).conj().T @ r_vec  # shape (r, 1)

        # Gradient with respect to tau:
        diag_elements = -2j * np.pi * (np.arange(N) - n) / N
        Lambda = np.diag(diag_elements)
    
        grad_tau = np.real(( (a_est.conj()) * (( Lambda @ G_mat @ Phi(tau_est)).conj().T) ) @ r_vec)

        

        # Preconditioner for tau-update:
        P_tau = np.diag(1/(-Kpp0) * ((a_est.flatten())**2))
        #P_tau = 1

        # Gradient descent update with learning rate eta
        #a_est = a_est - eta * grad_a #WE DONT MODIFY a_est because we suppose it is known 
        tau_est = tau_est - eta  * (P_tau @ grad_tau).flatten()

        tau1_traj.append(tau_est[0])
        tau2_traj.append(tau_est[1])

    print(P_tau)
    # Final estimated parameters (concatenated amplitudes and centroids)
    theta_est_final = np.concatenate((a_est.flatten(), tau_est))
    # True parameters vector: [weights, means]
    
    theta_star = np.concatenate((np.array(weights),np.array(means)))

    norm , theta_est_final_prime = permute_and_compute_norm(theta_est_final,theta_star,r)
    


    
    plt.plot(tau1_traj, tau2_traj, 'b.-', markersize=8, label="Tau running")  # Ligne bleue avec points
    # Marquer le point initial (blanc) et final (rouge)
    plt.scatter([tau1_traj[0]], [tau2_traj[0]], color='white', s=100, edgecolors='black', label="Tau est")
    plt.scatter([tau_star[0]], [tau_star[1]], color='red', s=100, edgecolors='black', label="Tau star")
    plt.legend()
    plt.show()

    return norm, theta_est_final , theta_est_final_prime







def show_conv_speed_tau(tau_est,tau_star,m,n,num_steps =1000,eta=0.1):
    """
    We take r=2, and let a1 = a2 = 1/2.

    We want to look at the landscape of the loss function L(teta) in this configuration.
    We Here do uniform sampling with N=2n+1 centered frequencies.

    Then, we do a gradient descent from theta_est to (we hope) theta_star.
    """
    assert len(tau_star) == 2 , "Issue : r !=2 "


    freqs=sample_uniform_frequencies(n)
    
    means = tau_star
    weights = [1/2,1/2]
    r = len(means)
    N=2*n+1

    #Useless here bcs we assume there is no noise
    """
    # Generate batch data and compute its FCE
    data = generate_data(m, means, weights)
    freqs = sample_uniform_frequencies(n)  # shape (N_freq, 1)
    baseline_freqs = sample_uniform_frequencies(n)
    """

    # Compute the empirical characteristic function once for the batch

    #Y = FCE(data, freqs)  # shape (N_freq, 1) #Noise coming from data generation

    
    # Initial ESPRIT estimation: theta_est = [a_est (r,), tau_est (r,)]

    ###theta_est = ESPRIT_(Y, r, fourier, baseline_freqs, freqs)
    #theta_est = np.array([1/2,1/2,4.9,5])
    #theta_est [0], theta_est[1] = 1/2,1/2 #ICI AUCUNE ESTIMATION SUR LES WEIGHTS

    a_est = np.array([1/2,1/2])
    a_star = np.array([1/2,1/2])
    tau_est = np.array(tau_est)

    theta_est = np.concatenate((a_est,tau_est))
    theta_star = np.concatenate((a_star,np.array(tau_star)))

    a_est = a_est.reshape(-1, 1)
    a_star = a_star.reshape(-1,1)

    vitesse_conv = [permute_and_compute_norm(theta_est,theta_star,r)[0]]



    # Precompute the "G" matrix: diagonal with entries given by kernel's Fourier transform at freqs.
    G_diag = fourier(freqs).flatten()  # shape (N_freq,)
    G_mat = np.diag(G_diag)  # shape (N_freq, N_freq)

    # Dictionary function Phi: for a given tau vector (r,), returns a matrix (N_freq x r)
    def Phi(tau):
        return np.exp(-2j * np.pi * np.outer(freqs.flatten(), tau))

    # Model function: Y_model = G * Phi(tau) * a.
    def Y_model(a, tau):
        return G_mat @ (Phi(tau) @ a)

    Y = Y_model(a_star,tau_star) # No noise assumed, perfect data m -> +inf

    # Compute K''(0): for K(t)=sum_f |ĝ(f)|^2 exp(2i*pi*f*t), we have
    # K''(0)= - (2*pi)^2 * sum_f |ĝ(f)|^2 * f^2.
    f_vals = freqs.flatten()
    Kpp0 = - (2 * np.pi)**2 * np.sum(np.abs(fourier(freqs).flatten())**2 * (f_vals**2))
 
    # Run batch gradient descent updates on the fixed FCE Y
    for step in range(num_steps):
        # Compute the current model prediction
        
        Y_estimated = Y_model(a_est, tau_est)
        # Residual: difference between model prediction and observed FCE
        r_vec = Y_estimated - Y  # shape (N_freq, 1)

        # Gradient with respect to amplitudes: grad_a = (G*Phi(tau))^H * r
        grad_a = (G_mat @ Phi(tau_est)).conj().T @ r_vec  # shape (r, 1)

        # Gradient with respect to tau:
        diag_elements = -2j * np.pi * (np.arange(N) - n) / N
        Lambda = np.diag(diag_elements)
    
        grad_tau = np.real(( (a_est.conj()) * (( Lambda @ G_mat @ Phi(tau_est)).conj().T) ) @ r_vec)

        # Preconditioner for tau-update:
        P_tau = np.diag(1/(-Kpp0) * ((a_est.flatten())**2))
        #P_tau = 1

        # Gradient descent update with learning rate eta
        #a_est = a_est - eta * grad_a #WE DONT MODIFY a_est because we suppose it is known 
        tau_est = tau_est - eta  * (P_tau @ grad_tau).flatten()

    
        theta_est = np.concatenate((a_est.flatten(),tau_est))

        vitesse_conv.append(permute_and_compute_norm(theta_est,theta_star,r)[0])

    
    # Final estimated parameters (concatenated amplitudes and centroids)
    theta_est_final = np.concatenate((a_est.flatten(), tau_est))
    # True parameters vector: [weights, means]
    
    theta_star = np.concatenate((np.array(weights),np.array(means)))

    norm , theta_est_final_prime = permute_and_compute_norm(theta_est_final,theta_star,r)
    


    plt.plot( vitesse_conv )
    plt.yscale('log')
    plt.xscale('log')
    plt.show()

    return norm, theta_est_final , theta_est_final_prime




def precond_diag(a_k,tau_k,freqs,r,n):
    """
    Build the adaptative diagonal preconditionner matrix.
    """
    K_second_0 = 0
 
    for f in freqs.flatten():
        K_second_0 += (abs(fourier(f))**2) * (f**2) * (-4*(np.pi**2))
    

    P = np.eye(2*r)

    for i in range(r):
        P[i+r,i+r] = (1/np.sqrt(-K_second_0) ) * (1/a_k[i])**2
        
    return P


def precond_full(a_k,tau_k,freqs,r,n):
    """
    Build the adaptative full preconditionner matrix.
    """
    N=2*n+1


    Ak = np.diag(np.concatenate((np.array([1 for _ in range(r)]), np.array(a_k))))
    

    G_diag = fourier(freqs).flatten()  # shape (N_freq,)
    G_mat = np.diag(G_diag)  # shape (N_freq, N_freq)

    Phi= np.exp(-2j * np.pi * np.outer(freqs.flatten(), tau_k))



    diag_elements = -2j * np.pi * (np.arange(N) - n) / N
    Lambda = np.diag(diag_elements)
    

    W = np.hstack((G_mat @ Phi , Lambda @ G_mat @ Phi))

    P = np.linalg.inv( Ak.conj().T @ ((W.conj().T) @ W) @ Ak)

    return P


def matrix_S(freqs,a_star,n,r):
    """
    Buils the matrix S that adimensionnate the space : With S, the convergence of the gradient descent is linear.
    """
    K_second_0 = 0
 
    for f in freqs.flatten():
        K_second_0 += (abs(fourier(f))**2) * (f**2) * (-4*(np.pi**2))
    
    S = np.diag(np.concatenate(([1/k for k in a_star] , [np.sqrt(-K_second_0) for _ in range(r)])))

    return S




def show_conv_speed(theta_est,theta_star,n,precond,num_steps =2000,eta=0.1):
    """
    We fix a preconditionner precond, and we look at the convergence of the gradient descent wrt to this preconditionner.
    We assume here the data is perfect : m is useless
    """
    assert len(theta_star) == len(theta_est) , "Error: star and est not same size"
    assert len(theta_star)%2 ==0 , "Error : r is not even"

    r = len(theta_star)//2
    freqs=sample_uniform_frequencies(n)

    #Make the input np.array
    theta_star = np.array(theta_star)
    theta_est = np.array(theta_est)

    tau_star = theta_star[r:]
    a_star = theta_star[:r]
    N=2*n+1

    tau_est = theta_est[r:]
    a_est = theta_est[:r]


    #Reshape the a vectors to make computations easier
    a_est = a_est.reshape(-1, 1)
    a_star = a_star.reshape(-1,1)

    S = matrix_S(freqs,a_star.flatten(),n,r)
    #vitesse_conv = [permute_and_compute_norm(theta_est,theta_star,r)[0]]

    vitesse_conv = [permute_and_compute_norm( (S @ theta_est.reshape(-1,1)).flatten() ,(S @ theta_est.reshape(-1,1)).flatten() ,r)[0]]

    # Precompute the "G" matrix: diagonal with entries given by kernel's Fourier transform at freqs.
    G_diag = fourier(freqs).flatten()  # shape (N_freq,)
    G_mat = np.diag(G_diag)  # shape (N_freq, N_freq)

    # Dictionary function Phi: for a given tau vector (r,), returns a matrix (N_freq x r)
    def Phi(tau):
        return np.exp(-2j * np.pi * np.outer(freqs.flatten(), tau))

    # Model function: Y_model = G * Phi(tau) * a.
    def Y_model(a, tau):
        return G_mat @ (Phi(tau) @ a)

    Y = Y_model(a_star,tau_star) # No noise assumed, perfect data m -> +inf
 
    # Run batch gradient descent updates 
    for step in range(num_steps):
        # Compute the current model prediction
        Y_estimated = Y_model(a_est, tau_est)

        # Residual: difference between model prediction and observed FCE
        r_vec = Y_estimated - Y  # shape (N_freq, 1)

        # Gradient with respect to amplitudes: grad_a = (G*Phi(tau))^H * r
        grad_a = (G_mat @ Phi(tau_est)).conj().T @ r_vec  # shape (r, 1)

        # Gradient with respect to tau:
        diag_elements = -2j * np.pi * (np.arange(N) - n) / N
        Lambda = np.diag(diag_elements)

        grad_tau = np.real(( (a_est.conj()) * (( Lambda @ G_mat @ Phi(tau_est)).conj().T) ) @ r_vec)

        #Concatenation of both gradients
        grad = np.vstack((grad_a,grad_tau))
        
        def Loss_theta(theta):
            return loss(theta,theta_star,r,freqs)

        # Gradient numérique si besoin
        #grad2 = nd.Gradient(Loss_theta)(theta_est) 

        #Preconditionner for the update
        P = precond(a_est.flatten(),tau_est,freqs,r,n)
       
        # Gradient descent update with learning rate eta    
        theta_est = theta_est - eta * (P @ grad).flatten()
        ##theta_est = theta_est -eta * grad.flatten()
        a_est = theta_est[:r].reshape(-1,1)
        tau_est = theta_est[r:]

        
        theta_est = np.concatenate((a_est.flatten(),tau_est))


        #vitesse_conv.append(permute_and_compute_norm(theta_est,theta_star,r)[0])
        vitesse_conv.append(permute_and_compute_norm( (S @ theta_est.reshape(-1,1)).flatten() ,(S @ theta_est.reshape(-1,1)).flatten() ,r)[0])
    
    
    # Final estimated parameters (concatenated amplitudes and centroids)
    theta_est_final = np.concatenate((a_est.flatten(), tau_est))

    norm , theta_est_final_prime = permute_and_compute_norm(theta_est_final,theta_star,r)
    

    print(vitesse_conv)
    plt.plot( vitesse_conv , label = f"{precond}")
    plt.yscale('log')
    plt.xlabel("Itération")
    plt.ylabel(f"Norme de la fonction de perte (échelle log)")
    plt.legend()
    plt.show()

    return norm, theta_est_final , theta_est_final_prime

